@@ -40,6 +40,7 @@
 #include <sys/stat.h>
 #include <stdio.h>
 #include <ctype.h>
+#include <limits.h>
 #define XK_LATIN1
 #include <X11/keysymdef.h>
 #include "Cv.h"
@@ -53,7 +54,7 @@ static Status LoadColornameDB(void);
  *		#define declarations local to this package.
  */
 #ifndef XCMSDB
-#define XCMSDB  "/usr/lib/X11/Xcms.txt"
+#define XCMSDB  XCMSDIR "/Xcms.txt"
 #endif
 
 #ifndef isgraph
@@ -125,13 +126,13 @@ _XcmsColorSpaceOfString(
  */
 {
     XcmsColorSpace	**papColorSpaces;
-    int n;
+    size_t n;
     char *pchar;
 
     if ((pchar = strchr(color_string, ':')) == NULL) {
 	return(XcmsFailure);
     }
-    n = (int)(pchar - color_string);
+    n = (size_t)(pchar - color_string);
 
     if (ccc == NULL) {
 	return(NULL);
@@ -197,7 +198,7 @@ _XcmsParseColorString(
     XcmsColorSpace	*pColorSpace;
     char		string_buf[64];
     char		*string_lowered;
-    int			len;
+    size_t		len;
     int			res;
 
     if (ccc == NULL) {
@@ -208,7 +209,9 @@ _XcmsParseColorString(
      * While copying color_string to string_lowered, convert to lowercase
      */
     if ((len = strlen(color_string)) >= sizeof(string_buf)) {
-	string_lowered = (char *) Xmalloc(len+1);
+	string_lowered = Xmalloc(len+1);
+	if (string_lowered == NULL)
+	    return(XcmsFailure);
     } else {
 	string_lowered = string_buf;
     }
@@ -254,7 +257,7 @@ FirstCmp(const void *p1, const void *p2)
  *
  */
 {
-    return(strcmp(((XcmsPair *)p1)->first, ((XcmsPair *)p2)->first));
+    return(strcmp(((const XcmsPair *)p1)->first, ((const XcmsPair *)p2)->first));
 }
 
 
@@ -313,7 +316,7 @@ field2(
 
     /* Find Field 1 */
     while (!isgraph(*pBuf)) {
-	if ((*pBuf != '\n') || (*pBuf != '\0')) {
+	if ((*pBuf == '\n') || (*pBuf == '\0')) {
 	    return(XcmsFailure);
 	}
 	if (isspace(*pBuf) || (*pBuf == delim)) {
@@ -416,8 +419,10 @@ _XcmsLookupColorName(
     tmpName = *name;
 
 Retry:
-    if ((len = strlen(tmpName)) > 63) {
-	name_lowered = (char *) Xmalloc(len+1);
+    if ((len = (int)strlen(tmpName)) > 63) {
+	name_lowered = Xmalloc(len+1);
+	if (name_lowered == NULL)
+	    return(XcmsFailure);
     } else {
 	name_lowered = name_lowered_64;
     }
@@ -504,7 +509,7 @@ RemoveSpaces(
 
     /* REMOVE SPACES */
     cptr = pString;
-    for (i = strlen(pString); i; i--, cptr++) {
+    for (i = (int)strlen(pString); i; i--, cptr++) {
 	if (!isspace(*cptr)) {
 	    *pString++ = *cptr;
 	    count++;
@@ -542,7 +547,10 @@ stringSectionSize(
     char *pBuf;
     char *f1;
     char *f2;
-    int i;
+    size_t i;
+
+    unsigned int numEntries = 0;
+    unsigned int sectionSize = 0;
 
     *pNumEntries = 0;
     *pSectionSize = 0;
@@ -576,26 +584,37 @@ stringSectionSize(
 	    return(XcmsFailure);
 	}
 
-	(*pNumEntries)++;
+	numEntries++;
+	if (numEntries >= INT_MAX)
+	    return(XcmsFailure);
 
-	(*pSectionSize) += (i = strlen(f1)) + 1;
+	i = strlen(f1);
+	if (i >= INT_MAX - sectionSize)
+	    return(XcmsFailure);
+	sectionSize += i + 1;
 	for (; i; i--, f1++) {
 	    /* REMOVE SPACES FROM COUNT */
 	    if (isspace(*f1)) {
-		(*pSectionSize)--;
+		sectionSize--;
 	    }
 	}
 
-	(*pSectionSize) += (i = strlen(f2)) + 1;
+	i = strlen(f2);
+	if (i >= INT_MAX - sectionSize)
+	    return(XcmsFailure);
+	sectionSize += i + 1;
 	for (; i; i--, f2++) {
 	    /* REMOVE SPACES FROM COUNT */
 	    if (isspace(*f2)) {
-		(*pSectionSize)--;
+		sectionSize--;
 	    }
 	}
 
     }
 
+    *pNumEntries = (int) numEntries;
+    *pSectionSize = (int) sectionSize;
+
     return(XcmsSuccess);
 }
 
@@ -715,11 +734,8 @@ LoadColornameDB(void)
     if ((pathname = getenv("XCMSDB")) == NULL) {
 	pathname = XCMSDB;
     }
-#ifdef __UNIXOS2__
-    pathname = __XOS2RedirRoot(pathname);
-#endif
 
-    length = strlen(pathname);
+    length = (int)strlen(pathname);
     if ((length == 0) || (length >= (BUFSIZ - 5))){
 	XcmsColorDbState = XcmsDbInitFailure;
 	return(XcmsFailure);
@@ -745,8 +761,15 @@ LoadColornameDB(void)
     }
     rewind(stream);
 
-    strings = (char *) Xmalloc(size);
-    pairs = (XcmsPair *)Xcalloc(nEntries, sizeof(XcmsPair));
+    strings = Xmalloc(size);
+    pairs = Xcalloc(nEntries, sizeof(XcmsPair));
+    if (strings == NULL || pairs == NULL) {
+	free(strings);
+	free(pairs);
+	(void) fclose(stream);
+	XcmsColorDbState = XcmsDbInitFailure;
+	return(XcmsFailure);
+    }
 
     ReadColornameDB(stream, pairs, strings);
     (void) fclose(stream);
